## Copyright (C) 2020 Armin Auf der Maur, 6230 Ebikon, Switzerland
## arminaufdermaur@yahoo.de
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <https://www.gnu.org/licenses/>.

## Author: Armin Auf der Maur <armin@LAPTOP-UL2NPON7>
## Revised: 2021-11-04

% Correlation Imposed Permutation CIP to calculate CI for bivariate regression
% Permutation with imposed correlation R for outcome y and treatment x.
% Permutation destroys correlations. It is re-established by
% a number of m1 pairs x and y sorted for maximum + or - correlation.
% m1 is calculated to obtain the observed (here) or any other possible R
% m0 = random pairs left in their original association, default =0
% set a value for m0a to select a specific m0 = m0a (condition m0a <n).
% set two values m0a<m0 and m0b>m0 to search m0 satisfying P(R|H0)=P(R=0|H1).
% R1dif2rr3 =1 (default) selects the calculations and graph for 1=R, 2=dif, 3=rr.
% A valid command to run H1CIP is e.g.: H1CIP("83E.txt","83sc.txt",3);

function retval = H1CIP (input1, input2, R1dif2rr3=1, m0a=0, m0b=m0a);
% This first command seeks two *.txt data files in the same directory.
% Input R1dif2rr3 selects: 1=R (Pearson), 2=dif (difference), 3=rr (risk ratio).
% A valid start typed into the command window to run the program is:
% H1CIP("83E.txt","83sc.txt");
% The semicolon after a command prevents printing. Important in loops!

NH0=100000;N =10000, % number of permutations for P(H0) and P(H1), respectively
% for more precision set N to 100000
PL = 0.025; PU = 0.025; % CI, also asymmetrical values possible

g=4; % number of runs determining m1 to obtain median(Ri) = R

rrmax =20; rrmin = 0.1; %setting upper amd lower bounds to the graph of rr

if (m0a ==m0b), stop =g; else, stop =3*g; endif, m0 =m0a;

y=load(input1); x=load(input2); n=length(y), ns=length(x(x!=0));

avy=mean(y);avx=mean(x);avxy=avy*avx;sty=std(y,1);stx=std(x,1);stxy=stx*sty;
% Terms of the Pearson R not changing by permutations.
% std(y, 1) is the uncorrected standard deviation (1/n, not 1/(n-1)).

Sxy=sum(y.*x); R=(Sxy/n-avxy)/stxy, cdif=sty*avx*n/(stx*ns);
% The dot in .* means a term by term multiplication of the data vectors y and x.
% cdif is a constant to calculate dif from the Pearson R: dif=R*cdif

dif=R*cdif; Rcr=avy*stx/(avx*sty); %R=Rcr is a singularity for rr

rr=1+(n/ns)/(-1+Rcr/R); CorTie=sign(R)*0.01/N; comp=Sxy-CorTie;
% The parameters dif and rr are associated with R by linear regression.
% CorTie decides for largest P in the case of ties.

nB4 = n /kurtosis(y), % A representative sample should have nB4 >=8

P=0; for(i=1:NH0), xperm=x(randperm(n)); Sxy=sum(y.*xperm);
if(Sxy>=comp), ++P; endif, endfor, PRH0=max(min(P/NH0, 1-P/NH0),1/(NH0-1)),
% This was to calculate the probability of the null hypothesis

% now preparations start to impose R for CI
y1=sort(y); x1=sort(x);y1x1=sum(y1.*x1);x1=sort(x,"descend");y2x2=sum(y1.*x1);
if(R>=0), Rmax=(y1x1/n-avxy)/(stxy), Rmin=(y2x2/n-avxy)/(stxy),
else, Rmax=(y2x2/n-avxy)/(stxy); Rmin=(y1x1/n-avxy)/(stxy); endif,
% Rmax is used for m1, Rmin in case of negative R.
  
m1 = R*(n -m0)/Rmax; % a first guess for m1

m1 = max(min(m1, n),0), % m1 should not be out of bounds;

DATA= complex(y, x); nnn=[1:n]; runs =0; m0,
% for the case of m0 >=1, a complex number is used for shuffling pairwise.

do, ++runs; % the outer do loop makes at least g runs to find m1
P=0; Sxy=0; Ri=zeros(1, N); mn=[1:(n-m0)]; m1Int=fix(m1);% m1 is not an integer

for (i=1:N); if(rem(i/m1, 1)>=rem(m1, 1)), mi=m1Int;
else, mi=1+m1Int; endif,
% m1 and m1+1 according to the residue rem(i/m1, 1) simulate m1

DAperm=DATA(randperm(n)); % permutation for random selection of pairs.

if (m0 ==0), x0y0=0; else, DA0=DAperm(nnn<=m0);y0=real(DA0);x0=imag(DA0);

x0y0=sum(y0.*x0); endif, % the contribution of m0

DA12=DAperm(nnn>m0);x12=imag(DA12);x12p=x12(randperm(n-m0));
% this permutation breaks the associations

x1=x12p(mn<=mi); x2=x12p(mn>mi); % x1 to be sorted to impose R, x2 permuted

if(R>=0); x1=sort(x1); else, x1=sort(x1, "descend"); endif,

y12 = real(DA12); y1=y12(mn<=mi); y1=sort(y1); y2=y12(mn>mi);

xy=(x0y0 +sum(y1.*x1)+sum(y2.*x2))/n; % The variable term xy: m0 +m1 +permuted

Sxy=Sxy+xy; if(xy>avxy), ++P; endif, % allows to calculate P(R=0|H1)

Ri(i)=(xy-avxy)/stxy; endfor, % collects N times Ri to form the cdf

P0H1=min(P/N, 1-P/N); % P(R=0|H1)

medR = median(Ri); meddif =medR *cdif; m1end =m1; % store some data

m1 = m1 +(R-medR)*(n -m0)/Rmax; % algorithm for a better m1

m1 = max(min(m1, n),0); % m1 should not go out of bounds

switch(runs), % here follows the algorithm to find m0 for P(R|H0)=P(R=0|H1)
case(1*g);  runs, m0a, m1=m1end, % just printing

Ri=sort(Ri); P0H1 = lookup(Ri, 0)/N; P0H1a=max(min(P0H1, 1-P0H1),1/N); m0 = m0b;

case (2*g), runs, m0b, m1=m1end, % to find d(P(H1))/d(m0)

Ri=sort(Ri); P0H1 = lookup(Ri, 0)/N; P0H1b=max(min(P0H1, 1-P0H1),1/N),

dP0H1dm0=(P0H1a-P0H1b)/(m0a-m0b), m0 =m0a+fix((PRH0-P0H1a)/dP0H1dm0),
% depending on the result other m0a or m0b could be selected for a next trial

endswitch,
%if default m0a=m0b, then stop is set at 1*g without executing further runs
until(runs>=stop), runs,
% Here ends the loop. Now comes printing and making a graph.

if (n.^(n-m1-m0)<=10*N), warning("n too small or R too close to Rmax"); endif,
Ri=sort(Ri); P0H1 = lookup(Ri, 0)/N; P0H1=max(min(P0H1, 1-P0H1),1/N),
meddif=cdif*median(Ri),

z =[1:N]; if(rem(N, 2)==0), z1=[1:N/2, N-[N/2:N-1]];
else, z1=[1:N/2+0.5, N-[0.5+N/2:N-1]]; endif, % these are the P values for Ri
 
switch(R1dif2rr3), 
% The switch-command selects and prepares the parameters for plotting.

case (1), % correlation R
medR=median(Ri), LR=Ri(z==N*PL), UR=Ri(z==(N*(1-PU))), 
if (R>=0), R16=Ri(z==N*0.159), else, R16=Ri(z==N*0.841), endif
 semilogy(Ri,z1/N,"b",R,PRH0,"xb",LR,PL,"ob",UR,PU,"ob",R16,0.16,"sb")
%semilogy(Ri,z1/N,"r",R,PRH0,"xr",LR,PL,"or",UR,PU,"or",R16,0.16,"sr")
%colours: b=blue, red, magenta, cyan, yellow, k=black
xlabel("R"),

case (2), dif =dif, % difference dif
Ldif=cdif*Ri(fix(N*PL)), Udif=cdif*Ri(fix(N*(1-PU))),dif16=cdif*Ri(fix(N*0.159))
if (R>=0),dif16=cdif*Ri(fix(N*0.159));else,dif16=cdif*Ri(fix(N*0.841)),endif,
if (min(x)<0),"warning: dif is for all x>=0",endif,c=cdif;L=Ldif;U=Udif;d=dif16;
semilogy(c*Ri,z1/N,"b",c*R, PRH0,"xb",L,PL,"ob",U,PU,"ob",d,0.16,"sb")
%semilogy(c*Ri,z1/N,"m",c*R,PRH0,"xm",L,PL,"om",U,PU,"om",d,0.16,"sm")
xlabel("dif"),

case (3), rr =rr, % ratio rr
Rc1 =0.9*Rcr, Rc2 =-0.9*Rcr*(ns/(n-ns)), rri=zeros(1,N);
for (i=1:N); rp(i)=max(min(Ri(i),Rc1),Rc2);
% Rc1 and Rc2 keep extreme permutations away from two singularities of rr.
rri(i)=max(min(1+(n/ns)/(-1+Rcr./rp(i)),rrmax),rrmin); endfor, 
rri=sort(rri); medrr = median(rri),
Lrr=rri(fix(N*PL)),Urr=rri(fix(N*(1-PU))), 
if (R>=0), rr16=rri(fix(N*0.159)), else, rr16=rri(fix(N*0.841)), endif
P=lookup(rri, 1)/N; Prr1H1=min(P, 1-P);
%semilogy(rri,z1/N,"b",rr,PRH0,"xb",Lrr,PL,"ob",Urr,PU,"ob",rr16,0.159,"sb")
loglog(rri,z1/N,"b",rr,PRH0,"xb",Lrr,PL,"ob",Urr,PU,"ob",rr16,0.159,"sb")
xlabel("rr"), endswitch, ylabel("P"), grid minor on, hold on,
display('Comment: PRH0=P(R|H0),  P0H1=P(Ri=0|H1), Ri=resampled permutations'),
endfunction